# oscillode
# Copyright (C) 2024 oscillode Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Provides the base protocol on which all ManagedObjs are derived, as well as explicit protocols describing common features supported by some `ManagedObjs`."""

import abc
import typing as typ

from blender_maxwell.utils import logger, serialize

from .. import contracts as ct

log = logger.get(__name__)


class ManagedObj(typ.Protocol):
	"""A weak name-based reference to some kind of object external to this software.

	While the object doesn't have to come from Blender's `bpy.types`, that is the driving motivation for this class: To encapsulate access to the powerful visual tools granted by Blender's 3D viewport, image editor, and UI.
	Through extensive trial-and-error, this transparently-cached immediate-mode interface has emerged as the best compromise.

	While not suited to all use cases, the `ManagedObj` paradigm is ideal for the common situation of 'node ownership of external logic'.
	For example, a plotting node needs access to its own image buffer; however, for it to be displayable, that image buffer must be owned by Blender.
	This is where `ManagedObj` provides the ability of the plotting node to "loosely own" a particular Blender image, which both allows the user to see/interact with the image as a typical Blender image, but also allows the node to drive ex. lifecycle (like removing the image plot when the node is deleted).

	This approach of loose name-coupling is not perfect. To name a few examples:

	- A particular challenge is that of identifying and namespacing managed datablocks. The user may, for example, change the name of an object - and what do we do then? Is generating new data under the now-vacant name a feature? Or is it a systematic memory leak backed by dangling fake-owned datablocks?
	- The precise definition of "loose ownership" is itself a matter of taste. Stronger ownership presumptions allow for stronger guarantees and simpler caching - but at the cost of more brittle breaks when the underlying objects are manipulated in ways we don't expect.
	- `.blend` persistance is not the default when it comes to certain Blender objects, which raises deeper UX questions about how opinionated we should be about where users put data.
	- Solving this doesn't help actually deal with data. There's a lot of very specific nuance in every single data pipeline, and that complexity is only added to watever this approach incurs.

	This abstract base class serves to provide a few of the most basic of commonly-available operations.
	In particular, implementations of `dump_as_msgspec`/`parse_as_msgspec` are enforced, for use with `oscillode.utils.serialize`.
	This way, ex. opening a `.blend` file will allow a newly-deserialized `ManagedObj` to re-attach transparently to the also-persisted underlying datablock.

	Parameters:
		managed_obj_type: Enum identifier indicating which of the `ct.ManagedObjType` the instance should declare itself as.
	"""

	managed_obj_type: ct.ManagedObjType

	@abc.abstractmethod
	def __init__(self, name: ct.ManagedObjName, prev_name: str | None = None):
		"""Initializes the managed object with a unique name.

		Use `prev_name` to indicate that the managed object will initially be avaiable under `prev_name`, but that it should be renamed to `name`.
		"""

	####################
	# - Properties
	####################
	@property
	@abc.abstractmethod
	def name(self) -> str:
		"""Retrieve the name of the managed object."""

	@name.setter
	@abc.abstractmethod
	def name(self, value: str) -> None:
		"""Retrieve the name of the managed object."""

	####################
	# - Methods
	####################
	@abc.abstractmethod
	def free(self) -> None:
		"""Cleanup the resources managed by the managed object."""

	@abc.abstractmethod
	def hide_preview(self) -> None:
		"""Hide any active preview of the managed object, if it exists, and if such an operation makes sense."""

	####################
	# - Serialization
	####################
	def dump_as_msgspec(self) -> serialize.NaiveRepresentation:
		"""Bijectively transform this managed object into a 'naive representation' that uses only basic Python types, which may be serialized cleanly."""
		return [
			serialize.TypeID.ManagedObj,
			self.__class__.__name__,
			self.name,
		]

	@staticmethod
	def parse_as_msgspec(obj: serialize.NaiveRepresentation) -> typ.Self:
		"""Bijectively construct an instance of this managed object from the 'naive representation', which may have been deserialized into."""
		return next(
			subclass(obj[2])
			for subclass in ManagedObj.__subclasses__()
			if subclass.__name__ == obj[1]
		)


####################
# - Support Flags
####################
class SupportsBlenderSelect(typ.Protocol):
	"""Protocol guaranteeing the ability to select the object in Blender."""

	def bl_select(self) -> None:
		"""Select the managed object in Blender."""


class Supportsreview(typ.Protocol):
	"""Protocol guaranteeing support for previewing the object."""

	def show_preview(self) -> None:
		"""Shows a preview of the managed object."""

	def hide_preview(self) -> None:
		"""Hide any active preview of the managed object, if it exists."""
